import gym
import numpy as np
import random
import matplotlib.pyplot as plt
from collections import deque
from itertools import combinations
from collections import defaultdict
import cv2
import copy
import os
from zqmtool.other import remove_make_path, get_row_index_in_arr, get_rgba_from_cmap

random.seed(0)
np.random.seed(0)

wall = np.array([
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
    [1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1],
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
], dtype=int)

init_pos = [5,5]

train_goal_states = np.asarray([
    [0, 0], [3, 0], [0, 3], [3, 3],
    [0, 7], [0, 10], [3, 7], [3, 10],
    [10, 0], [10, 3], [7, 0], [7, 3],
    [7, 10], [10, 10], [7, 7], [10, 7]
])

room_bound = {0: np.asarray([[0, 3], [0, 3]]),
              1: np.asarray([[0, 3], [7, 10]]),
              2: np.asarray([[7, 10], [0, 3]]),
              3: np.asarray([[7, 10], [7, 10]])}

# wall = np.array([
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1],
#     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
#     [1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
#     [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
# ], dtype=int)
# # plt.imshow(wall)
# # plt.show()
#
#
# train_goal_states = np.asarray([
#     [0, 0], [4, 0], [0, 4], [4, 4],
#     [0, 8], [0, 12], [4, 8], [4, 12],
#     [12, 0], [12, 4], [8, 0], [8, 4],
#     [8, 12], [12, 12], [8, 8], [12, 8]
# ])
#
# room_bound = {0: np.asarray([[0, 4], [0, 4]]),
#               1: np.asarray([[0, 4], [8, 12]]),
#               2: np.asarray([[8, 12], [0, 4]]),
#               3: np.asarray([[8, 12], [8, 12]])}


def gen_uninformative_goal_from_goal_state(goal_state, goal_id, num_goal):
    # # wh = max(wall.shape)
    # wh = 10
    # rnd = wh * np.random.random(size=(wh,))
    # # rnd_id = random.choice(range(0,9))
    # rnd_id = int(wh / 2)
    # rnd[rnd_id:rnd_id + 2] = goal_state.copy()
    # goal = rnd.copy()

    import torch
    import torch.nn as nn
    torch.manual_seed(0)

    encoder = nn.Embedding(num_embeddings=num_goal, embedding_dim=10)
    with torch.no_grad():
        goal = encoder(torch.LongTensor([goal_id]))[0].numpy().tolist()
    return goal


def get_goal_dataset(wall):
    goal_info_dict = {}
    goal_state_to_goal_id = defaultdict(dict)
    goal_id = -1
    goal_set = []

    state_set = np.stack(np.where(wall == 0), axis=-1)

    for state_id, state in enumerate(state_set):
        goal_state = state.copy()

        cur_room_id = None
        for room_id, bounds in room_bound.items():
            x_bound, y_bound = bounds
            if (goal_state[0] in range(x_bound[0], x_bound[1] + 1)) and (
                    goal_state[1] in range(y_bound[0], y_bound[1] + 1)):
                cur_room_id = room_id
                break
        if cur_room_id is None:
            continue

        goal_id += 1

        goal = gen_uninformative_goal_from_goal_state(goal_state, goal_id, num_goal=len(state_set))

        goal_set.append(goal)

        mode = 'train' if get_row_index_in_arr(goal_state, train_goal_states) is not False else 'test'

        goal_state_to_goal_id[goal_state[0]][goal_state[1]] = goal_id

        goal_info_dict[goal_id] = {'goal': goal, 'mode': mode, 'goal_state': goal_state, 'room_id': cur_room_id}

    goal_set = np.asarray(goal_set)

    return goal_set, state_set, goal_info_dict, goal_state_to_goal_id


goal_set, state_set, goal_info_dict, goal_state_to_goal_id = get_goal_dataset(wall)

positive_goal_id_pair = set()


class Env(gym.Env):

    def __init__(self, mode, random_initial_state, can_see_test_goal=True):
        super().__init__()
        self.can_see_test_goal = can_see_test_goal
        self.mode = mode
        self.height, self.width = wall.shape

        self.actions = [
            np.array([0, 1]),  # Up
            np.array([1, 0]),  # Right
            np.array([0, -1]),  # Down
            np.array([-1, 0]),  # Left,
        ]

        self.goal_dataset = {goal_id: goal_info for goal_id, goal_info in goal_info_dict.items() if
                             goal_info['mode'] == self.mode}

        self.state_dim = state_set.shape[-1]
        self.goal_dim = len(list(self.goal_dataset.values())[0]['goal'])

        self.random_initial_state = random_initial_state if random_initial_state is not None else \
            {'train': True, 'test': False}[mode]

        self.T = 100  # if random_initial_state == True else 500

        self.action_space = gym.spaces.Discrete(n=len(self.actions))
        self.observation_space = gym.spaces.Box(low=-np.inf, high=np.inf, shape=(2 + self.goal_dim,))

    def vis_goal_state_pair(self, ):
        logdir = 'vis_gpal_state_pair'
        remove_make_path(logdir)
        for pair in positive_goal_id_pair:
            g_id_0, g_id_1 = pair

            if not os.path.exists(f'{g_id_0} {g_id_1}.png'):
                g_s_0 = self.goal_dataset[g_id_0]['goal_state']
                g_s_1 = self.goal_dataset[g_id_1]['goal_state']
                wall_copy = wall.copy()
                wall_copy[g_s_0[0], g_s_0[1]] = 2
                wall_copy[g_s_1[0], g_s_1[1]] = 2
                plt.imshow(wall_copy)
                plt.savefig(os.path.join(logdir, f'{g_id_0} {g_id_1}.png'))
                plt.close()

    def check_success(self, ):
        return (self.pos[0] == self.goal_state[0]) and (self.pos[1] == self.goal_state[1])

    def update_state(self, a):

        def f():
            if (self.pos[0] in goal_state_to_goal_id.keys()) and \
                    (self.pos[1] in goal_state_to_goal_id[self.pos[0]].keys()):
                self.achieved_goal_id = goal_state_to_goal_id[self.pos[0]][self.pos[1]]
                if (not self.can_see_test_goal) and (goal_info_dict[self.achieved_goal_id]['mode'] == 'test'):
                    self.achieved_goal_id = -1
            else:
                self.achieved_goal_id = -1
            self.time_window.append(self.achieved_goal_id)
            self.update_positive_goal_id_pair()

        if a is not None:
            a = self.actions[a]
            tmp_pos = self.pos.copy() + a
            x, y = tmp_pos
            if (x >= 0 and x < self.height) and (y >= 0 and y < self.width) and (wall[x, y] == 0):
                self.pos = tmp_pos
                self.pos_id = get_row_index_in_arr(self.pos, state_set)

        f()

        self.cur_goal_id = -1
        if (self.achieved_goal_id >= 0) and (self.achieved_goal_id not in self.achieved_goal_id_set):
            self.cur_goal_id = copy.deepcopy(self.achieved_goal_id)
            self.achieved_goal_id_set.add(self.achieved_goal_id)

    def update_positive_goal_id_pair(self, ):
        time_window_no_neg = [elem for elem in set(self.time_window) if elem >= 0]
        adjecent_goal_id_pairs = set(combinations(time_window_no_neg, 2))
        positive_goal_id_pair.update(adjecent_goal_id_pairs)

    def get_mask(self):
        mask = []
        for act in self.actions:
            tmp_pos = self.pos.copy() + act
            x, y = tmp_pos
            if (x >= 0 and x < self.height) and (y >= 0 and y < self.width) and (wall[x, y] == 0):
                mask.append(True)
            else:
                mask.append(False)
        return mask

    def get_obs(self):
        return {'state_id': self.pos_id, 'cur_goal_id': self.cur_goal_id + 1, 'goal_id': self.goal_id,
                'hindsight_goal_id': self.achieved_goal_id,
                'mask': self.get_mask(),
                'max_rew': 10., 'mode': self.mode}

    def reset(self, **kwargs):
        self.t = 0

        self.achieved_goal_id = None
        self.achieved_goal_id_set = set()

        self.time_window = deque(maxlen=10)

        self.goal_id = random.choice(list(self.goal_dataset.keys()))
        self.goal_state, self.goal = self.goal_dataset[self.goal_id]['goal_state'], \
                                     self.goal_dataset[self.goal_id]['goal']

        self.pos = state_set[random.choice(range(len(state_set)))] \
            if self.random_initial_state else np.asarray(init_pos)
        self.pos_id = get_row_index_in_arr(self.pos, state_set)

        self.update_state(a=None)

        return self.get_obs()

    def step(self, action):
        self.t += 1

        self.update_state(action)
        success = self.check_success()

        done = (self.t == self.T) or (success)

        rew = 10 * float(success) - 1 / self.T + float(self.cur_goal_id > 0)

        return self.get_obs(), rew, done, {}


def make_env(mode, random_initial_state=None):
    return Env(mode, random_initial_state)


env = make_env(mode='train')

output_for_goal_representation_learning = {
    'wall': wall,
    'g_dim': env.goal_dim,
    'state_dim': np.prod(env.state_dim),
    'act_dim': env.action_space.n,
    'state_set': state_set,
    'goal_set': goal_set,
    'adjacent_goal_id_pairs': positive_goal_id_pair,
    'goal_info_dict': goal_info_dict,
    'c_dict': {k: get_rgba_from_cmap('Set2', k) for k in range(4)},
}

if __name__ == '__main__':
    env = Env(mode='train', random_initial_state=True)

    for i in range(1):

        s = env.reset()
        d = False
        wall_copy = wall.copy()
        while not d:
            pos = state_set[env.pos_id]
            wall_copy[pos[0], pos[1]] = 2
            s, r, d, _ = env.step(env.action_space.sample())



import time
from collections import defaultdict
from typing import Callable, Dict, Optional, Union

import numpy as np
import tqdm

from tianshou.data import Collector
from tianshou.policy import BasePolicy
from tianshou.trainer import gather_info
from tianshou.utils import BaseLogger, LazyLogger, MovAvg, tqdm_config

from typing import Any, Callable, Dict, Optional, Union


def test_episode(
        policy: BasePolicy,
        collector: Collector,
        test_fn: Optional[Callable[[int, Optional[int]], None]],
        epoch: int,
        n_episode: int,
        logger: Optional[BaseLogger] = None,
        global_step: Optional[int] = None,
        reward_metric: Optional[Callable[[np.ndarray], np.ndarray]] = None,
) -> Dict[str, Any]:
    """A simple wrapper of testing policy in collector."""
    collector.reset_env()
    collector.reset_buffer()
    policy.eval()
    if test_fn:
        test_fn(epoch, global_step)
    result = collector.collect(n_episode=n_episode)
    if reward_metric:
        result["rews"] = reward_metric(result["rews"])
    if logger and global_step is not None:
        logger.log_test_data(result, global_step)
    return result


def offpolicy_trainer(
        policy: BasePolicy,
        train_collector: Collector,
        test_collector: Collector,
        max_epoch: int,
        step_per_epoch: int,
        step_per_collect: int,
        episode_per_test: int,
        batch_size: int,
        update_per_step: Union[int, float] = 1,
        train_fn: Optional[Callable[[int, int], None]] = None,
        test_fn: Optional[Callable[[int, Optional[int]], None]] = None,
        stop_fn: Optional[Callable[[float], bool]] = None,
        save_fn: Optional[Callable[[BasePolicy], None]] = None,
        save_checkpoint_fn: Optional[Callable[[int, int, int], None]] = None,
        resume_from_log: bool = False,
        reward_metric: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        logger: BaseLogger = LazyLogger(),
        verbose: bool = True,
        test_in_train: bool = True,
) -> Dict[str, Union[float, str]]:
    """A wrapper for off-policy trainer procedure.

    The "step" in trainer means an environment step (a.k.a. transition).

    :param policy: an instance of the :class:`~tianshou.policy.BasePolicy` class.
    :param Collector train_collector: the collector used for training.
    :param Collector test_collector: the collector used for testing.
    :param int max_epoch: the maximum number of epochs for training. The training
        process might be finished before reaching ``max_epoch`` if ``stop_fn`` is set.
    :param int step_per_epoch: the number of transitions collected per epoch.
    :param int step_per_collect: the number of transitions the collector would collect
        before the network update, i.e., trainer will collect "step_per_collect"
        transitions and do some policy network update repeatedly in each epoch.
    :param episode_per_test: the number of episodes for one policy evaluation.
    :param int batch_size: the batch size of sample data, which is going to feed in the
        policy network.
    :param int/float update_per_step: the number of times the policy network would be
        updated per transition after (step_per_collect) transitions are collected,
        e.g., if update_per_step set to 0.3, and step_per_collect is 256, policy will
        be updated round(256 * 0.3 = 76.8) = 77 times after 256 transitions are
        collected by the collector. Default to 1.
    :param function train_fn: a hook called at the beginning of training in each epoch.
        It can be used to perform custom additional operations, with the signature ``f(
        num_epoch: int, step_idx: int) -> None``.
    :param function test_fn: a hook called at the beginning of testing in each epoch.
        It can be used to perform custom additional operations, with the signature ``f(
        num_epoch: int, step_idx: int) -> None``.
    :param function save_fn: a hook called when the undiscounted average mean reward in
        evaluation phase gets better, with the signature ``f(policy: BasePolicy) ->
        None``.
    :param function save_checkpoint_fn: a function to save training process, with the
        signature ``f(epoch: int, env_step: int, gradient_step: int) -> None``; you can
        save whatever you want.
    :param bool resume_from_log: resume env_step/gradient_step and other metadata from
        existing tensorboard log. Default to False.
    :param function stop_fn: a function with signature ``f(mean_rewards: float) ->
        bool``, receives the average undiscounted returns of the testing result,
        returns a boolean which indicates whether reaching the goal.
    :param function reward_metric: a function with signature ``f(rewards: np.ndarray
        with shape (num_episode, agent_num)) -> np.ndarray with shape (num_episode,)``,
        used in multi-agent RL. We need to return a single scalar for each episode's
        result to monitor training in the multi-agent RL setting. This function
        specifies what is the desired metric, e.g., the reward of agent 1 or the
        average reward over all agents.
    :param BaseLogger logger: A logger that logs statistics during
        training/testing/updating. Default to a logger that doesn't log anything.
    :param bool verbose: whether to print the information. Default to True.
    :param bool test_in_train: whether to test in the training phase. Default to True.

    :return: See :func:`~tianshou.trainer.gather_info`.
    """
    start_epoch, env_step, gradient_step = 0, 0, 0
    if resume_from_log:
        start_epoch, env_step, gradient_step = logger.restore_data()
    last_rew, last_len = 0.0, 0
    stat: Dict[str, MovAvg] = defaultdict(MovAvg)
    start_time = time.time()
    train_collector.reset_stat()
    test_collector.reset_stat()
    test_in_train = test_in_train and train_collector.policy == policy
    # test_result = test_episode(
    #     policy, test_collector, test_fn, start_epoch, episode_per_test, logger,
    #     env_step, reward_metric
    # )
    test_fn(start_epoch, env_step)

    for epoch in range(1 + start_epoch, 1 + max_epoch):
        # train
        policy.train()
        with tqdm.tqdm(
                total=step_per_epoch, desc=f"Epoch #{epoch}", **tqdm_config
        ) as t:
            while t.n < t.total:
                if train_fn:
                    train_fn(epoch, env_step)
                result = train_collector.collect(n_step=step_per_collect)
                if result["n/ep"] > 0 and reward_metric:
                    result["rews"] = reward_metric(result["rews"])
                env_step += int(result["n/st"])
                t.update(result["n/st"])
                logger.log_train_data(result, env_step)
                last_rew = result['rew'] if 'rew' in result else last_rew
                last_len = result['len'] if 'len' in result else last_len
                data = {
                    "env_step": str(env_step),
                    "rew": f"{last_rew:.2f}",
                    "len": str(int(last_len)),
                    "n/ep": str(int(result["n/ep"])),
                    "n/st": str(int(result["n/st"])),
                }
                if result["n/ep"] > 0:
                    if test_in_train and stop_fn and stop_fn(result["rew"]):
                        test_result = test_episode(
                            policy, test_collector, test_fn, epoch, episode_per_test,
                            logger, env_step
                        )
                        if stop_fn(test_result["rew"]):
                            if save_fn:
                                save_fn(policy)
                            logger.save_data(
                                epoch, env_step, gradient_step, save_checkpoint_fn
                            )
                            t.set_postfix(**data)
                            return gather_info(
                                start_time, train_collector, test_collector,
                                test_result["rew"], test_result["rew_std"]
                            )
                        else:
                            policy.train()
                for _ in range(round(update_per_step * result["n/st"])):
                    gradient_step += 1
                    losses = policy.update(batch_size, train_collector.buffer)
                    for k in losses.keys():
                        stat[k].add(losses[k])
                        losses[k] = stat[k].get()
                        data[k] = f"{losses[k]:.3f}"
                    logger.log_update_data(losses, gradient_step)
                    t.set_postfix(**data)
            if t.n <= t.total:
                t.update()
        # test
        test_fn(epoch, env_step)

        logger.save_data(epoch, env_step, gradient_step, save_checkpoint_fn)

    return gather_info(
        start_time, train_collector, test_collector, 0, 0
    )


import os, sys

sys.path.append(os.path.dirname(__file__))

from copy import deepcopy
from typing import Any, Dict, Optional, Union

from tianshou.data import Batch, ReplayBuffer, to_numpy, to_torch_as
from tianshou.policy import DQNPolicy

import torch
from tianshou.utils.net.common import MLP
import torch.nn as nn
import numpy as np
from torch.nn.functional import softplus
from torch.distributions import Normal, Independent
import torch.nn.functional as F
import math
import random
from itertools import combinations
from other import make_path_if_not_exist
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
import seaborn as sns


def update_beta(method, beta, loss_infomin=None, acc=None, min_val=None):
    def add_beta(beta):
        return min(beta + step_val, max_val)

    def reduce_beta(beta):
        return max(beta - step_val, min_val)

    max_val = 1
    min_val = 0.001 if min_val is None else min_val
    step_val = 1e-3

    if method == 'InfoGoal':
        return add_beta(beta) if acc > 0.95 else reduce_beta(beta)
    elif method == 'InfoBot':
        return 0.01
    else:
        return beta


def reparameterization(params, z_dim, num_sample=1):
    mu, sigma = params[:, :z_dim], params[:, z_dim:]
    sigma = softplus(sigma) + 1e-3  # Make sigma always positive
    if num_sample == 1:
        z = Independent(Normal(loc=mu, scale=sigma), 1).rsample()
    else:
        z = Independent(Normal(loc=mu, scale=sigma), 1).rsample(sample_shape=(num_sample,))
    return z, mu, sigma

def infomax_loss_fn_two_goal_repre(z0,z1):
    return  F.l1_loss(z0, z1, reduction='none')

def infomax_loss_fn(z_p0, z_p1, z_n0, z_n1):
    z_pos = infomax_loss_fn_two_goal_repre(z_p0,z_p1)
    z_neg = infomax_loss_fn_two_goal_repre(z_n0,z_n1)

    z_pos = torch.mean(z_pos, dim=-1)
    z_neg = torch.mean(z_neg, dim=-1)
    loss = torch.exp(z_pos - z_neg)
    loss = torch.mean(loss)

    acc = torch.mean(torch.as_tensor(z_pos < z_neg, dtype=torch.float32))
    return loss, acc


def infomin_loss_fn(mu, std, reduction='mean'):
    # loss = F.mse_loss(std, torch.ones_like(std))+F.mse_loss(mu, torch.zeros_like(mu))
    if reduction == 'mean':
        loss = -0.5 * (1 + 2 * std.log() - mu.pow(2) - std.pow(2)).sum(1).mean().div(math.log(2))
    elif reduction == 'none':
        loss = -0.5 * (1 + 2 * std.log() - mu.pow(2) - std.pow(2)).sum(1).div(math.log(2))
    return loss


def draw_scatter_emb(net, save_path, env_info, env, times, device, show_goal_id_set=None, enlarge_goal_id=None):
    make_path_if_not_exist(os.path.dirname(save_path))

    SMALL_SIZE = 6 * times
    MEDIUM_SIZE = 8 * times

    plt.rc('pdf', fonttype=42)
    plt.rc('ps', fonttype=42)
    plt.rc('font', family='serif')
    plt.rc('font', size=MEDIUM_SIZE)  # controls default text sizes
    plt.rc('axes', titlesize=MEDIUM_SIZE)  # fontsize of the axes title
    plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
    plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
    plt.rc('figure', titlesize=MEDIUM_SIZE)  # fontsize of the figure title

    goal_info_dict = env_info['goal_info_dict']
    c_dict = env_info['c_dict']

    cur_goal_ids = []
    cur_goals = []
    cur_colors = []
    cur_modes = []
    cur_sizes = []
    for cur_goal_id in goal_info_dict.keys():
        if show_goal_id_set is not None:
            if cur_goal_id not in show_goal_id_set:
                continue
        cur_mode = 'training' if cur_goal_id in env.training_goal_ids else 'test'
        if (enlarge_goal_id is not None) and (cur_mode == 1) and (cur_goal_id != enlarge_goal_id): continue
        cur_modes.append(cur_mode)
        cur_goal_ids.append(cur_goal_id)
        cur_colors.append(goal_info_dict[cur_goal_id]['room'])
        if cur_goal_id == enlarge_goal_id:
            cur_sizes.append(400)
        else:
            cur_sizes.append(200)

    with torch.no_grad():
        z, m, s = net.encode_fn(cur_goal_ids, type='goal', num_sample=100)
        z_orig_shape = z.shape
        if len(z_orig_shape) == 2:
            z_orig_shape = (1, *z_orig_shape)
        num_repeat = z_orig_shape[0]
        num_goal = z_orig_shape[1]
        num_dim = z_orig_shape[2]

        cur_colors_for_z = cur_colors*num_repeat #[cur_color for cur_color in cur_colors for i in range(num_repeat)]
        cur_modes_for_z = cur_modes*num_repeat #[cur_mode for cur_mode in cur_modes for i in range(num_repeat)]
        cur_sizes_for_z = cur_sizes*num_repeat #[cur_size for cur_size in cur_sizes for i in range(num_repeat)]

        cur_colors_for_m = cur_colors
        cur_modes_for_m = cur_modes
        cur_sizes_for_m = cur_sizes

        # cur_colors_for_z = np.asarray(cur_colors)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)
        # cur_modes_for_z = np.asarray(cur_modes)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)
        # cur_sizes_for_z = np.asarray(cur_sizes)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)

        # cur_colors_for_m = np.asarray(cur_colors).reshape(-1, 1)
        # cur_modes_for_m = np.asarray(cur_modes).reshape(-1, 1)
        # cur_sizes_for_m = np.asarray(cur_sizes).reshape(-1,1)

        z = z.reshape(-1, num_dim).cpu().numpy()
        z_2d = z if num_dim == 2 else PCA(n_components=2).fit_transform(z)
        m_2d = np.mean(z_2d.reshape(z_orig_shape[0], z_orig_shape[1], -1), axis=0)

        z_2d_w_color = [[z2i[0], z2i[1], cur_color_for_z, cur_mode_for_z, cur_size_for_z] for
                        z2i, cur_color_for_z, cur_mode_for_z, cur_size_for_z in
                        zip(z_2d, cur_colors_for_z, cur_modes_for_z, cur_sizes_for_z)]
        m_2d_w_color = [[m2i[0], m2i[1], cur_color_for_m, cur_mode_for_m, cur_size_for_m] for
                        m2i, cur_color_for_m, cur_mode_for_m, cur_size_for_m in
                        zip(m_2d, cur_colors_for_m, cur_modes_for_m, cur_sizes_for_m)]


        # z_2d_w_color = np.concatenate([z_2d, cur_colors_for_z, cur_modes_for_z, cur_sizes_for_z], axis=-1)
        # m_2d_w_color = np.concatenate([m_2d, cur_colors_for_m, cur_modes_for_m, cur_sizes_for_m], axis=-1)

    z_pd = pd.DataFrame(z_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])
    m_pd = pd.DataFrame(m_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])

    # m_pd = m_pd.replace({'mode': {0: 'training', 1: 'test'}})
    # m_pd = m_pd.replace({'room': {0: 'A', 1: 'B', 2: 'C', 3: 'D'}})

    ax = plt.gca()
    sns.scatterplot(ax=ax, data=z_pd, x='x', y='y',
                    alpha=0.1, style='mode', markers={'training': 'o', 'test': '*'}, s=400,
                    hue='room', palette=c_dict, legend=False)
    sns.scatterplot(ax=ax, data=m_pd, x='x', y='y',
                    style='mode', markers={'training': 'o', 'test': '*'}, s=400,
                    hue='room', palette=c_dict, legend=True)

    # for label, (marker, s) in {'training goal': ('o', 200), 'test goal': ('*', 400)}.items():
    #     plt.scatter([], [], s=s, marker=marker, label=label, facecolors='none', edgecolors='k')
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), shadow=True, markerscale=2)  # ,

    # ax.axis('off')
    # ax.set_aspect('auto')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xlabel('')
    ax.set_ylabel('')
    plt.savefig(save_path, bbox_inches='tight')
    plt.close()

class Base:

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        self.net = net
        self.optim = optim
        self.device = net.device
        self.batch_size = batch_size
        self.positive_instance_pairs = positive_instance_pairs
        self.len_positive_instance_pairs = None
        self.pos_pair_set = None
        self.neg_pair_set = None

    def divide_pos_neg_pairs(self, ):
        if (self.len_positive_instance_pairs is not None) and \
                (len(self.positive_instance_pairs) == self.len_positive_instance_pairs):
            return

        self.len_positive_instance_pairs = len(self.positive_instance_pairs)

        # get z_a
        a = []
        for pair in self.positive_instance_pairs:
            a += list(pair)
        a = set(a)

        pos_pair_set = []
        for ai in a:
            pos = []
            for pair in self.positive_instance_pairs:
                if ai in pair:
                    pos += list(pair)
            pos = set(pos)

            for posi in pos:
                pos_pair_set.append([ai, posi])

        neg_pair_set = []
        for pair in combinations(a, 2):
            is_pos = False
            for pos_pair in pos_pair_set:
                if ((pair[0] == pos_pair[0]) and (pair[1] == pos_pair[1])) or (pair[1] == pos_pair[0]) and (
                        pair[0] == pos_pair[1]):
                    is_pos = True
                    break

            if not is_pos:
                neg_pair_set.append(pair)

        self.pos_pair_set = pos_pair_set
        self.neg_pair_set = neg_pair_set

    def forward(self):
        raise NotImplementedError

    def update(self):
        self.optim.zero_grad()
        loss, loss_breakdown = self.forward()
        loss.backward()
        self.optim.step()

        desc = " ".join(['{} {:.3f}'.format(k, v) for k, v in loss_breakdown.items()])
        return desc


class InfoMax(Base):

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        super().__init__(
            net=net, optim=optim, batch_size=batch_size,
            positive_instance_pairs=positive_instance_pairs
        )

    def forward(self):
        self.divide_pos_neg_pairs()
        if (self.pos_pair_set is None) or \
                ((self.neg_pair_set is None)) or \
                (min(len(self.neg_pair_set), len(self.pos_pair_set)) == 0):
            return {'loss': None}

        num_sample = min(len(self.neg_pair_set), len(self.pos_pair_set))
        pos_pairs = np.asarray(random.sample(self.pos_pair_set, num_sample))
        neg_pairs = np.asarray(random.sample(self.neg_pair_set, num_sample))

        a = torch.as_tensor(self.net.goal_set[pos_pairs[:, 0]], dtype=torch.float32, device=self.net.device)
        b = torch.as_tensor(self.net.goal_set[pos_pairs[:, 1]], dtype=torch.float32, device=self.net.device)
        c = torch.as_tensor(self.net.goal_set[neg_pairs[:, 0]], dtype=torch.float32, device=self.net.device)
        d = torch.as_tensor(self.net.goal_set[neg_pairs[:, 1]], dtype=torch.float32, device=self.net.device)

        abcd = torch.cat([a, b, c, d])
        z_abcd, mu, std = self.net.encode_goal(abcd)
        z_a, z_b, z_c, z_d = torch.chunk(z_abcd, 4)

        loss_infomax, infomax_acc = infomax_loss_fn(z_p0=z_a, z_p1=z_b, z_n0=z_c, z_n1=z_d)

        return {'loss': loss_infomax, 'acc': infomax_acc,
                'mu': mu, 'std': std}


class InfoMin(Base):

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        super().__init__(
            net=net, optim=optim, batch_size=batch_size,
            positive_instance_pairs=positive_instance_pairs
        )

    def forward(self, mu=None, std=None):

        if mu is None:
            pair = np.asarray(list(self.positive_instance_pairs)).copy()
            np.random.shuffle(pair)
            anc = pair[:self.batch_size, 0]
            pos = pair[:self.batch_size, 1]

            if random.random() < 0.5:
                tmp = anc.copy()
                anc = pos.copy()
                pos = tmp

            anc = torch.as_tensor(self.net.goal_set[anc], dtype=torch.float32,
                                  device=self.net.device)
            pos = torch.as_tensor(self.net.goal_set[pos], dtype=torch.float32,
                                  device=self.net.device)

            z_a, m_a, s_a = self.net.encode_goal(anc)
            with torch.no_grad():
                z_pos, m_p, s_p = self.net.encode_goal(pos)

            mu = torch.cat([m_a, m_p])
            std = torch.cat([s_a, s_p])

        loss = infomin_loss_fn(mu, std)
        return {'loss': loss}



import copy

import imageio
import cv2
import numpy as np
import os.path
import os
import imageio
import cv2
import open3d as o3d
import scipy.io as sio
import json
import cv2
import random
import PIL.Image
import torch
import networkx as nx
from functools import reduce
from PIL import Image
from numpy import asarray
from ast import literal_eval
from matplotlib import pyplot as plt
from numpy.linalg import inv
from tqdm import tqdm
from glob import glob
from zqmtool.image import to_rgb

# from open3d import JVisualizer
# from open3d.j_visualizer import JVisualizer


"""GENERAL FUNCTIONS """


def load_point_clouds(voxel_size, path):
    pcd = o3d.io.read_point_cloud(path)
    pcd_down = pcd.voxel_down_sample(voxel_size)
    return pcd_down


def getCamera():
    return o3d.camera.PinholeCameraIntrinsic(
        o3d.camera.PinholeCameraIntrinsicParameters.Kinect2ColorCameraDefault)


#     # Get camera information for AVD scene
#     if os.path.isfile(datapath + scene + "/cameras.txt"):
#         f = open(datapath + scene + "/cameras.txt")
#     else: # if the scene does not have camera file then just use the one from scene 001_1
#         f = open(datapath + "Home_002_1/cameras.txt")

#     data = f.readlines()
#     tok = data[-1].split(" ")
#     intr = []

#     for i in range(4,len(tok)):
#         intr.append(float(tok[i]))

#     return np.asarray(intr, dtype=np.float32) # fx, fy, cx, cy, distortion params


def load_scene_info(avd_root):
    # annotations contains the action information
    # image_structs contains camera information for each image
    image_structs_path = os.path.join(avd_root, 'image_structs.mat')
    annotation_path = os.path.join(avd_root, 'annotations.json')
    data = sio.loadmat(image_structs_path)

    scale = data['scale'][0][0]
    image_structs = data['image_structs'][0]

    rot = image_structs['R']
    trans = image_structs['t']
    im_names_all = image_structs['image_name']  # info 0 # list of image names in the scene
    im_names_all = np.hstack(im_names_all)  # flatten the array
    world_poses = image_structs['world_pos']  # info 3
    directions = image_structs['direction']  # info 4
    annotations = json.load(open(annotation_path))
    return trans, rot, scale, im_names_all


def load_scene_info_rot(avd_root, scene):
    # annotations contains the action information
    # image_structs contains camera information for each image
    image_structs_path = os.path.join(avd_root, scene, 'image_structs.mat')
    annotation_path = os.path.join(avd_root, scene, 'annotations.json')
    data = sio.loadmat(image_structs_path)

    scale = data['scale'][0][0]
    image_structs = data['image_structs'][0]

    rot = image_structs['R']
    trans = image_structs['t']
    im_names_all = image_structs['image_name']  # info 0 # list of image names in the scene
    im_names_all = np.hstack(im_names_all)  # flatten the array
    world_poses = image_structs['world_pos']  # info 3
    directions = image_structs['direction']  # info 4
    annotations = json.load(open(annotation_path))
    return data, world_poses, directions


""" PROJECT RELATED FUNCTIONS """


# visualize the mask on any category (normal, Depth, RGB), for any image (rot)
# filetype , type of the output filr (jpg,png)
# outputFile, the name of the folder of output
# read_dictionary, directory that stors the mask.npy

def maskVisualization(read_dictionary, category, rot, filetype, outputFolder):
    masks = read_dictionary[rot[-19:-4] + '.jpg']
    count = 0

    for j in masks:
        print(j)

        # out = np.zeros((1080,1920,3))
        # new_im = cv2.imread(category +'/'+ rot[-19:-4]+filetype)
        # out[:,:,0] = np.where(j==True,new_im[:,:,0], 255)
        # out[:,:,1] = np.where(j==True,new_im[:,:,1], 255)
        # out[:,:,2] = np.where(j==True,new_im[:,:,2], 255)

        # if os.path.isfile("{}/{}_{}_{}{}".format(outputFolder,rot[-19:-4],category, count,filetype)):
        # print("True")
        # imageio.imwrite("{}/{}_{}_{}{}".format(outputFolder, rot[-19:-4],category, count, filetype),out.astype(np.uint8))

        # count = count+1


def addImgsTOgather(imgPath, outFolder):
    for filename1 in os.listdir(imgPath):
        idd = 0
        for filename2 in os.listdir(imgPath):

            if (filename1[0:15] == filename2[0:15] and idd == 0):
                img = cv2.imread(imgPath + '/' + filename2)
                idd = 1

            elif (filename1[0:15] == filename2[0:15] and idd == 1):
                img = cv2.imread(imgPath + '/' + filename2) + img

        cv2.imwrite("{}/{}{}".format(outFolder, filename1[0:15], filename1[-4:]), img)
        # imageio.imwrite("{}/{}{}".format(outFolder, filename1[0:15],filename1[-4:]),img)


def generatePcdinCameraCoordinat(imgPath, depthpath):
    # im = cv2.imread(imgPath,cv2.IMREAD_UNCHANGED)
    # newsize = (320,240)
    # im1 = cv2.resize(im, newsize)
    # Final_img = im1.astype(dtype=np.float32)
    # cv2.imwrite("out.jpg", Final_img)

    intr = getCamera()
    intr.width = 1920
    intr.height = 1080
    intrinsic_params = intr
    # intrinsic_params = o3d.camera.PinholeCameraIntrinsic()
    # intrinsic_params.set_intrinsics(1920, 1080, intr[0], intr[1], intr[2], intr[3])
    # intrinsic_params.set_intrinsics(320,240,intr[0]/6, intr[1]/4.5, intr[2]/6, intr[3]/4.5)

    color_img = o3d.io.read_image(imgPath)
    # print("color_img",imgPath,color_img)

    depth_img = o3d.io.read_image(depthpath)
    # print('depth_img',depthpath,depth_img)

    rgbd_img = o3d.geometry.RGBDImage.create_from_color_and_depth(color_img, depth_img, convert_rgb_to_intensity=False)
    pcd = o3d.geometry.PointCloud.create_from_rgbd_image(rgbd_img, intrinsic_params)
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])

    return pcd




def convertPcd2WorldCoordinat(pcd, datapath, imgName):
    if imgName.find('.jpg')==-1:
        imgName = imgName+'.jpg'
    voxel_size = 0.02
    pcd = pcd.voxel_down_sample(voxel_size=voxel_size)
    pcd.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30))
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])

    TRS = load_scene_info(datapath)

    [j], = np.where(TRS[3] == imgName)

    R = inv(TRS[1][j])
    t = np.dot(-R, TRS[0][j]) * TRS[2] / 1000
    # print("R Matrix : " , R.shape[0])
    # print("T Matrix : " , t.shape[0])

    if (R.shape[0] != 3 and t.shape[0] != 3):
        return None, None
    else:
        Transformation_Matrix = [[R[0][0], R[0][1], R[0][2], t[0]],
                                 [R[1][0], R[1][1], R[1][2], t[1]],
                                 [R[2][0], R[2][1], R[2][2], t[2]],
                                 [0, 0, 0, 1]]

        pcd = o3d.geometry.PointCloud.transform(pcd, Transformation_Matrix)
        return pcd, Transformation_Matrix


def multiViewPCL(listPclPath, pcdMultiViewOutFolder, name="outMultiView"):
    voxel_size = 0.02

    flag = 0
    for i in listPclPath:
        print("i", i)
        if (flag == 0):
            pcd = load_point_clouds(voxel_size, "{}".format(i))
            flag = 1
        else:
            pcdNew = load_point_clouds(voxel_size, "{}".format(i))
            pcd = pcd + pcdNew

    o3d.io.write_point_cloud("{}/{}.ply".format(pcdMultiViewOutFolder, name), pcd)
    # print(name + ".ply is saved.")


def mapPcdToGround(Pcdpath, mapOutFolder):
    voxel_size = 0.02
    pcd = load_point_clouds(voxel_size, "{}".format(Pcdpath))

    color_cloud = np.asarray(pcd.colors)
    Points = np.asarray(pcd.points)

    plt.scatter(Points[:, 0], Points[:, 2], c=color_cloud, s=1)
    plt.savefig('{}/{}'.format(mapOutFolder, Pcdpath))


def load_coordinate(path, shape, color):
    Transformation_Matrix = np.load(path,allow_pickle=True)
    if shape == 'box':
        scale = 0.2
        coordinate = o3d.geometry.TriangleMesh.create_box(width=scale, height=scale, depth=scale)
    elif shape == 'arrow':
        scale = 0.6
        coordinate = o3d.geometry.TriangleMesh.create_arrow(
            cylinder_height=1e-3, cylinder_radius=1e-3,
            cone_height=scale,cone_radius=0.1,
        )
    else:
        raise NotImplementedError
    coordinate.paint_uniform_color(color)
    # coordinate = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.5)
    coordinate = o3d.geometry.PointCloud.transform(coordinate, Transformation_Matrix)
    return coordinate


def merge_pcd(path, colors, pcdOutFolder, init_pcd_list=[]):
    voxel_size = 0.02
    pcd_list = copy.deepcopy(init_pcd_list)
    for i, img_name in tqdm(enumerate(glob(os.path.join(pcdOutFolder, '*.ply')))):
        img_name = os.path.basename(img_name).split('.')[0]
        if len(init_pcd_list) == 0:
            pcd_i = load_point_clouds(voxel_size, os.path.join(pcdOutFolder, f'{img_name}.ply'))
            pcd_i.transform([[1, 0, 0, 0], [0, 0, -1, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            pcd_list.append(pcd_i)


        if (img_name + '.jpg' in path):
            shape = 'arrow'
            coordinate_i = load_coordinate(os.path.join(pcdOutFolder, img_name+'.npy'),shape, colors[path.index(img_name + '.jpg')])
            coordinate_i.transform([[1, 0, 0, 0], [0, 0, -1, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            pcd_list.append(coordinate_i)

    return pcd_list


def visualize_pcd_list(pcd_list, vis, show=True, img_size=2048):
    vis.clear_geometries()
    for pcd in pcd_list:
        vis.add_geometry(pcd)
    set_visualizer_view(vis)

    if show:
        vis.run()
        vis.destroy_window()
        vis.close()
    else:
        vis.poll_events()
        vis.update_renderer()
        state = np.asarray(vis.capture_screen_float_buffer(do_render=True))
        h, w, c = state.shape
        state = cv2.resize(state, (int(img_size * (w / h)), img_size))
        state = np.asarray(state * 255, dtype=np.uint8)
        state = to_rgb(state)
        return state




def define_visualizer(visible):
    vis = o3d.visualization.Visualizer()
    vis.create_window(visible=visible)

    renderopt = vis.get_render_option()
    renderopt.point_size = 0.1

    return vis

def set_visualizer_view(vis):
    ctr = vis.get_view_control()
    # ctr.change_field_of_view(step=-60)
    ctr.set_zoom(0.45)
    # ctr.rotate(0, -100, 0)



from copy import deepcopy
from typing import Any, Dict, Optional, Union

from tianshou.data import Batch, to_torch_as
from tianshou.policy import DQNPolicy

import torch
from tianshou.utils.net.common import MLP
import torch.nn as nn
import numpy as np


class Policy(DQNPolicy):
    """Implementation of Deep Q Network. arXiv:1312.5602.

    Implementation of Double Q-Learning. arXiv:1509.06461.

    Implementation of Dueling DQN. arXiv:1511.06581 (the dueling DQN is
    implemented in the network side, not here).

    :param torch.nn.Module model: a model following the rules in
        :class:`~tianshou.policy.BasePolicy`. (s -> logits)
    :param torch.optim.Optimizer optim: a torch.optim for optimizing the model.
    :param float discount_factor: in [0, 1].
    :param int estimation_step: the number of steps to look ahead. Default to 1.
    :param int target_update_freq: the target network update frequency (0 if
        you do not use the target network). Default to 0.
    :param bool reward_normalization: normalize the reward to Normal(0, 1).
        Default to False.
    :param bool is_double: use double dqn. Default to True.

    .. seealso::

        Please refer to :class:`~tianshou.policy.BasePolicy` for more detailed
        explanation.
    """

    def __init__(
            self,
            goal_representation_learner,
            state_representation_learner,
            model: torch.nn.Module,
            optim: torch.optim.Optimizer,
            discount_factor: float = 0.99,
            estimation_step: int = 1,
            target_update_freq: int = 0,
            reward_normalization: bool = False,
            is_double: bool = True,
            **kwargs: Any,
    ) -> None:
        super().__init__(model, optim, discount_factor, estimation_step,
                         target_update_freq, reward_normalization, is_double, **kwargs)
        self.goal_representation_learner = goal_representation_learner
        self.state_representation_learner = state_representation_learner

    def learn(self, batch: Batch, **kwargs: Any) -> Dict[str, float]:
        loss_breakdown = {}
        if self._target and self._iter % self._freq == 0:
            self.sync_weight()
        self.optim.zero_grad()
        weight = batch.pop("weight", 1.0)
        q = self(batch).logits
        q = q[np.arange(len(q)), batch.act]
        r = to_torch_as(batch.returns.flatten(), q)
        td = r - q
        loss = (td.pow(2) * weight).mean()
        batch.weight = td  # prio-buffer

        loss_breakdown.update({'rl': loss.item()})

        loss_g, loss_g_breakdown = self.goal_representation_learner.forward(batch=batch)
        loss_s, loss_s_breakdown = self.state_representation_learner.forward(batch=batch)

        if loss_g is not None:
            loss_breakdown.update(loss_g_breakdown)
            loss = loss + loss_g

        if loss_s is not None:
            loss_breakdown.update(loss_s_breakdown)
            loss = loss + loss_s

        loss.backward()

        self.optim.step()

        self._iter += 1

        return loss_breakdown



import io
from enum import Enum
from typing import List, Optional, Union
from typing import Dict, List, Tuple

import numpy as np
import onnxruntime as ort
from PIL import Image
from PIL.Image import Image as PILImage

import hashlib
import os
import sys
from contextlib import redirect_stdout
from pathlib import Path
from typing import Type

import gdown

from cv2 import (
    BORDER_DEFAULT,
    MORPH_ELLIPSE,
    MORPH_OPEN,
    GaussianBlur,
    getStructuringElement,
    morphologyEx,
)

from pymatting.alpha.estimate_alpha_cf import estimate_alpha_cf
from pymatting.foreground.estimate_foreground_ml import estimate_foreground_ml
from pymatting.util.util import stack_images
from scipy.ndimage import binary_erosion

from scipy.special import log_softmax

from typing import List

from PIL import Image
from PIL.Image import Image as PILImage

kernel = getStructuringElement(MORPH_ELLIPSE, (3, 3))

pallete1 = [
    0,
    0,
    0,
    255,
    255,
    255,
    0,
    0,
    0,
    0,
    0,
    0,
]

pallete2 = [
    0,
    0,
    0,
    0,
    0,
    0,
    255,
    255,
    255,
    0,
    0,
    0,
]

pallete3 = [
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    255,
    255,
    255,
]


class BaseSession:
    def __init__(self, model_name: str, inner_session: ort.InferenceSession):
        self.model_name = model_name
        self.inner_session = inner_session

    def normalize(
            self,
            img: PILImage,
            mean: Tuple[float, float, float],
            std: Tuple[float, float, float],
            size: Tuple[int, int],
    ) -> Dict[str, np.ndarray]:
        im = img.convert("RGB").resize(size, Image.Resampling.LANCZOS)

        im_ary = np.array(im)
        im_ary = im_ary / np.max(im_ary)

        tmpImg = np.zeros((im_ary.shape[0], im_ary.shape[1], 3))
        tmpImg[:, :, 0] = (im_ary[:, :, 0] - mean[0]) / std[0]
        tmpImg[:, :, 1] = (im_ary[:, :, 1] - mean[1]) / std[1]
        tmpImg[:, :, 2] = (im_ary[:, :, 2] - mean[2]) / std[2]

        tmpImg = tmpImg.transpose((2, 0, 1))

        return {
            self.inner_session.get_inputs()[0]
                .name: np.expand_dims(tmpImg, 0)
                .astype(np.float32)
        }

    def predict(self, img: PILImage) -> List[PILImage]:
        raise NotImplementedError


class SimpleSession(BaseSession):
    def predict(self, img: PILImage) -> List[PILImage]:
        ort_outs = self.inner_session.run(
            None,
            self.normalize(
                img, (0.485, 0.456, 0.406), (0.229, 0.224, 0.225), (320, 320)
            ),
        )

        pred = ort_outs[0][:, 0, :, :]

        ma = np.max(pred)
        mi = np.min(pred)

        pred = (pred - mi) / (ma - mi)
        pred = np.squeeze(pred)

        mask = Image.fromarray((pred * 255).astype("uint8"), mode="L")
        mask = mask.resize(img.size, Image.Resampling.LANCZOS)

        return [mask]


class ClothSession(BaseSession):
    def predict(self, img: PILImage) -> List[PILImage]:
        ort_outs = self.inner_session.run(
            None, self.normalize(img, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), (768, 768))
        )

        pred = ort_outs
        pred = log_softmax(pred[0], 1)
        pred = np.argmax(pred, axis=1, keepdims=True)
        pred = np.squeeze(pred, 0)
        pred = np.squeeze(pred, 0)

        mask = Image.fromarray(pred.astype("uint8"), mode="L")
        mask = mask.resize(img.size, Image.LANCZOS)

        masks = []

        mask1 = mask.copy()
        mask1.putpalette(pallete1)
        mask1 = mask1.convert("RGB").convert("L")
        masks.append(mask1)

        mask2 = mask.copy()
        mask2.putpalette(pallete2)
        mask2 = mask2.convert("RGB").convert("L")
        masks.append(mask2)

        mask3 = mask.copy()
        mask3.putpalette(pallete3)
        mask3 = mask3.convert("RGB").convert("L")
        masks.append(mask3)

        return masks


class ReturnType(Enum):
    BYTES = 0
    PILLOW = 1
    NDARRAY = 2


def new_session(model_name: str) -> BaseSession:
    session_class: Type[BaseSession]

    if model_name == "u2netp":
        md5 = "8e83ca70e441ab06c318d82300c84806"
        url = "https://drive.google.com/uc?id=1tNuFmLv0TSNDjYIkjEdeH1IWKQdUA4HR"
        session_class = SimpleSession
    elif model_name == "u2net":
        md5 = "60024c5c889badc19c04ad937298a77b"
        url = "https://drive.google.com/uc?id=1tCU5MM1LhRgGou5OpmpjBQbSrYIUoYab"
        session_class = SimpleSession
    elif model_name == "u2net_human_seg":
        md5 = "c09ddc2e0104f800e3e1bb4652583d1f"
        url = "https://drive.google.com/uc?id=1ZfqwVxu-1XWC1xU1GHIP-FM_Knd_AX5j"
        session_class = SimpleSession
    elif model_name == "u2net_cloth_seg":
        md5 = "2434d1f3cb744e0e49386c906e5a08bb"
        url = "https://drive.google.com/uc?id=15rKbQSXQzrKCQurUjZFg8HqzZad8bcyz"
        session_class = ClothSession
    else:
        assert AssertionError(
            "Choose between u2net, u2netp, u2net_human_seg or u2net_cloth_seg"
        )

    home = os.getenv("U2NET_HOME", os.path.join("~", ".u2net"))
    path = Path(home).expanduser() / f"{model_name}.onnx"
    path.parents[0].mkdir(parents=True, exist_ok=True)

    if not path.exists():
        with redirect_stdout(sys.stderr):
            gdown.download(url, str(path), use_cookies=False)
    else:
        hashing = hashlib.new("md5", path.read_bytes(), usedforsecurity=False)
        if hashing.hexdigest() != md5:
            with redirect_stdout(sys.stderr):
                gdown.download(url, str(path), use_cookies=False)

    sess_opts = ort.SessionOptions()

    if "OMP_NUM_THREADS" in os.environ:
        sess_opts.inter_op_num_threads = int(os.environ["OMP_NUM_THREADS"])

    return session_class(
        model_name,
        ort.InferenceSession(
            str(path), providers=ort.get_available_providers(), sess_options=sess_opts
        ),
    )


def alpha_matting_cutout(
        img: PILImage,
        mask: PILImage,
        foreground_threshold: int,
        background_threshold: int,
        erode_structure_size: int,
) -> PILImage:
    if img.mode == "RGBA" or img.mode == "CMYK":
        img = img.convert("RGB")

    img = np.asarray(img)
    mask = np.asarray(mask)

    is_foreground = mask > foreground_threshold
    is_background = mask < background_threshold

    structure = None
    if erode_structure_size > 0:
        structure = np.ones(
            (erode_structure_size, erode_structure_size), dtype=np.uint8
        )

    is_foreground = binary_erosion(is_foreground, structure=structure)
    is_background = binary_erosion(is_background, structure=structure, border_value=1)

    trimap = np.full(mask.shape, dtype=np.uint8, fill_value=128)
    trimap[is_foreground] = 255
    trimap[is_background] = 0

    img_normalized = img / 255.0
    trimap_normalized = trimap / 255.0

    alpha = estimate_alpha_cf(img_normalized, trimap_normalized)
    foreground = estimate_foreground_ml(img_normalized, alpha)
    cutout = stack_images(foreground, alpha)

    cutout = np.clip(cutout * 255, 0, 255).astype(np.uint8)
    cutout = Image.fromarray(cutout)

    return cutout


def naive_cutout(img: PILImage, mask: PILImage) -> PILImage:
    empty = Image.new("RGBA", (img.size), 0)
    cutout = Image.composite(img, empty, mask)
    return cutout


def get_concat_v_multi(imgs: List[PILImage]) -> PILImage:
    pivot = imgs.pop(0)
    for im in imgs:
        pivot = get_concat_v(pivot, im)
    return pivot


def get_concat_v(img1: PILImage, img2: PILImage) -> PILImage:
    dst = Image.new("RGBA", (img1.width, img1.height + img2.height))
    dst.paste(img1, (0, 0))
    dst.paste(img2, (0, img1.height))
    return dst


def post_process(mask: np.ndarray) -> np.ndarray:
    """
    Post Process the mask for a smooth boundary by applying Morphological Operations
    Research based on paper: https://www.sciencedirect.com/science/article/pii/S2352914821000757
    args:
        mask: Binary Numpy Mask
    """
    mask = morphologyEx(mask, MORPH_OPEN, kernel)
    mask = GaussianBlur(mask, (5, 5), sigmaX=2, sigmaY=2, borderType=BORDER_DEFAULT)
    mask = np.where(mask < 127, 0, 255).astype(np.uint8)  # convert again to binary
    return mask


def remove(
        data: Union[bytes, PILImage, np.ndarray],
        alpha_matting: bool = False,
        alpha_matting_foreground_threshold: int = 240,
        alpha_matting_background_threshold: int = 10,
        alpha_matting_erode_size: int = 10,
        session: Optional[BaseSession] = None,
        only_mask: bool = False,
        post_process_mask: bool = False,
) -> Union[bytes, PILImage, np.ndarray]:
    if isinstance(data, PILImage):
        return_type = ReturnType.PILLOW
        img = data
    elif isinstance(data, bytes):
        return_type = ReturnType.BYTES
        img = Image.open(io.BytesIO(data))
    elif isinstance(data, np.ndarray):
        return_type = ReturnType.NDARRAY
        img = Image.fromarray(data)
    else:
        raise ValueError("Input type {} is not supported.".format(type(data)))

    if session is None:
        session = new_session("u2net")

    masks = session.predict(img)
    cutouts = []

    for mask in masks:
        if post_process_mask:
            mask = Image.fromarray(post_process(np.array(mask)))

        if only_mask:
            cutout = mask

        elif alpha_matting:
            try:
                cutout = alpha_matting_cutout(
                    img,
                    mask,
                    alpha_matting_foreground_threshold,
                    alpha_matting_background_threshold,
                    alpha_matting_erode_size,
                )
            except ValueError:
                cutout = naive_cutout(img, mask)

        else:
            cutout = naive_cutout(img, mask)

        cutouts.append(cutout)

    cutout = img
    if len(cutouts) > 0:
        cutout = get_concat_v_multi(cutouts)

    if ReturnType.PILLOW == return_type:
        return cutout

    if ReturnType.NDARRAY == return_type:
        return np.asarray(cutout)

    bio = io.BytesIO()
    cutout.save(bio, "PNG")
    bio.seek(0)

    return bio.read()



import setuptools

with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="zqmtool",
    version='0.0.62',
    author='zou_qiming',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=False,
    packages=setuptools.find_packages(),
)



import shutil

import torch
import matplotlib.pyplot as plt
import numpy as np
import os
import seaborn as sns
import cv2

from collections import defaultdict
import matplotlib.patches as patches
from open3d.visualization import Visualizer
from zqmtool.image import save_image, to_rgb
import open3d as o3d
import time
from zqmtool.multi_view_stereo import merge_pcd, visualize_pcd_list
from zqmtool.other import remove_make_path


def save_traj(collector, output_for_goal_representation_learning):
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']
    wall = output_for_goal_representation_learning['wall']
    batch, _ = collector.buffer.sample(batch_size=0)
    s_ids = batch.obs_next.state_id
    g_ids = batch.obs_next.goal_id
    ds = batch.done

    traj_infos = defaultdict(dict)
    for s_id, g_id, d_i in zip(s_ids, g_ids, ds):
        s = output_for_goal_representation_learning['state_set'][s_id]
        if g_id not in traj_infos.keys():
            traj_infos[g_id]['goal_state'] = None
            traj_infos[g_id]['state_visitation'] = np.zeros_like(wall)
        traj_infos[g_id]['goal_state'] = goal_id_to_goal_info[g_id]['goal_state']
        traj_infos[g_id]['state_visitation'][s[0], s[1]] += 1

    return traj_infos


def save_traj_2(collector, output_for_goal_representation_learning):
    state_name_set = output_for_goal_representation_learning['state_name_set']
    state_name_achieved_goal_id = output_for_goal_representation_learning['state_name_achieved_goal_id']
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']
    batch, _ = collector.buffer.sample(batch_size=0)
    s_ids = batch.obs_next.state_id
    g_ids = batch.obs_next.goal_id
    ds = batch.done

    traj_infos = defaultdict(dict)
    tmp = []
    t = 0
    for s_id, g_id, d_i in zip(s_ids, g_ids, ds):
        if g_id in tmp: continue
        t += 1
        if g_id not in traj_infos.keys():
            traj_infos[g_id]['goal_state'] = None
            traj_infos[g_id]['visited_state_names'] = []
            traj_infos[g_id]['hit_g'] = defaultdict(list)
        traj_infos[g_id]['goal_state'] = goal_id_to_goal_info[g_id]['goal_state_names']
        traj_infos[g_id]['visited_state_names'].append(state_name_set[s_id])
        for hit_g_id in state_name_achieved_goal_id[state_name_set[s_id]]:
            traj_infos[g_id]['hit_g'][hit_g_id].append({'t': t, 'state_name': state_name_set[s_id]})

        if d_i and (g_id not in tmp):
            t = 0
            tmp.append(g_id)

    for g_id in traj_infos.keys():
        traj_infos[g_id]['visited_state_names'] = list(dict.fromkeys(traj_infos[g_id]['visited_state_names']))

    return traj_infos


def draw_traj_2(annotations_dir, orig_state_dir, mode, collector, output_for_goal_representation_learning,
                test_env, args, vis, pcdOutFolder, enlarge_goal_id=None):
    c_dict = output_for_goal_representation_learning['c_dict']
    state_set = output_for_goal_representation_learning['state_set']
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']
    state_name_pcd_indicator = output_for_goal_representation_learning['state_name_pcd_indicator']
    goalId2objId = output_for_goal_representation_learning['goalId2objId']
    room_obj_id_map = output_for_goal_representation_learning['room_obj_id_map']

    traj_info = save_traj_2(collector=collector,
                            output_for_goal_representation_learning=output_for_goal_representation_learning)
    goal_ids = {'training': test_env.training_goal_ids, 'test': test_env.test_goal_ids}[mode]

    init_pcd_list = merge_pcd([], colors=None, pcdOutFolder=pcdOutFolder)
    for cur_goal_id in goal_ids:
        print(traj_info[cur_goal_id]['hit_g'])

        # if goalId2objId[cur_goal_id] not in room_obj_id_map['kitchen']: continue

        if (enlarge_goal_id is not None) and (mode == 'test') and (cur_goal_id != enlarge_goal_id): continue

        c = c_dict[goal_id_to_goal_info[cur_goal_id]['room']]
        marker = {'training': 'o', 'test': '*'}[mode]
        img_size = 2048
        colors = []
        achieved_goal_state_names = [elem['state_name'] for list in traj_info[cur_goal_id]['hit_g'].values() for elem in
                                     list]
        for indice, state_name in enumerate(traj_info[cur_goal_id]['visited_state_names']):
            if state_name in achieved_goal_state_names:
                colors.append([0, 1, 0])
                if (indice == len(traj_info[cur_goal_id]['visited_state_names']) - 1) and (
                        cur_goal_id in traj_info[cur_goal_id]['hit_g'].keys()):
                    colors[-1] = [1, 0, 0]
            else:
                colors.append([0, 0, 1])
        pcd_list = merge_pcd(traj_info[cur_goal_id]['visited_state_names'], init_pcd_list=init_pcd_list, colors=colors,
                             pcdOutFolder=pcdOutFolder)

        state = visualize_pcd_list(pcd_list, vis, show=False)
        # state = state[169:1953, 425:2069, :]

        out_dir = os.path.join(args.logdir,
                               f"traj/{mode} {int(cur_goal_id in traj_info[cur_goal_id]['hit_g'].keys())} objID {goalId2objId[cur_goal_id]} num hit goal {len(traj_info[cur_goal_id]['hit_g'])}")
        remove_make_path(out_dir)
        save_image(os.path.join(out_dir, 'slam.png'), state)

        for hit_g_id in traj_info[cur_goal_id]['hit_g'].keys():
            for hit_goal_info in traj_info[cur_goal_id]['hit_g'][hit_g_id]:
                state_name = hit_goal_info['state_name']
                t = hit_goal_info['t']
                aux_title = 'achievedObj' if hit_g_id != cur_goal_id else 'desiredObj'
                src_orig_state_dir = os.path.join(orig_state_dir, state_name)
                dst_state_dir = os.path.join(out_dir, '{} {}'.format(aux_title, state_name))
                shutil.copy(src=src_orig_state_dir, dst=dst_state_dir)


def draw_traj(epoch, mode, collector, output_for_goal_representation_learning, test_env, args, enlarge_goal_id=None):
    def draw_training_goal_state(ax):
        for training_goal_id in [elem for elem in goal_id_to_goal_info.keys() if
                                 goal_id_to_goal_info[elem]['mode'] == 'train']:
            c = c_dict[goal_id_to_goal_info[training_goal_id]['room_id']]
            marker = {'train': 'o', 'test': '*'}['train']
            g_s = goal_id_to_goal_info[training_goal_id]['goal_state']
            ax.scatter(g_s[1] + 0.5, g_s[0] + 0.5, marker=marker, c=c, s=200)
        return ax

    c_dict = output_for_goal_representation_learning['c_dict']
    state_set = output_for_goal_representation_learning['state_set']
    goal_id_to_goal_info = output_for_goal_representation_learning['goal_info_dict']

    traj_info = save_traj(collector=collector,
                          output_for_goal_representation_learning=output_for_goal_representation_learning)

    for cur_goal_id in [elem for elem in goal_id_to_goal_info.keys() if goal_id_to_goal_info[elem]['mode'] == mode]:

        if (enlarge_goal_id is not None) and (mode == 'test') and (cur_goal_id != enlarge_goal_id): continue

        c = c_dict[goal_id_to_goal_info[cur_goal_id]['room_id']]
        marker = {'train': 'o', 'test': '*'}[mode]

        ax = plt.gca()
        ax.axis('off')
        # ax.set_aspect('auto')

        heatmap = sns.heatmap(ax=ax, data=traj_info[cur_goal_id]['state_visitation'], cmap='OrRd')
        cbar = heatmap.collections[0].colorbar
        cbar.set_ticks([])
        cbar.set_label('state visitation count for test goal')

        obs = test_env.reset()
        s_id = obs['state_id']
        init_s = state_set[s_id]
        ax.scatter(init_s[0] + 0.5, init_s[1] + 0.5, marker='s', c='b', s=200)
        for goal_id in traj_info.keys():
            if goal_id != cur_goal_id: continue
            g_s = traj_info[goal_id]['goal_state']
            ax.scatter(g_s[1] + 0.5, g_s[0] + 0.5, marker=marker, c=c, s=400)

        for block_pos in np.stack(np.where(wall == 1), axis=-1):
            ax.add_patch(
                patches.Rectangle(xy=(block_pos[1], block_pos[0]), width=1, height=1, facecolor='k', fill=True))

        ax = draw_training_goal_state(ax)

        # for label, (marker, s) in {'training goal': ('o', 200), 'test goal': ('*', 400)}.items():
        #     plt.scatter([], [], s=s, marker=marker, label=label, facecolors='none', edgecolors='k')
        # plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.01), shadow=True, ncol=2)  # , markerscale=3

        plt.savefig(os.path.join(args.logdir, f'{epoch}_{mode}_{args.rl}_{args.method}_{cur_goal_id}.png'),
                    bbox_inches='tight')
        plt.close()
